package sets;


public class LazySet<T> extends OptimisticSet<T>
{
    public LazySet() {
        super();
    }

    // are `pred' and `curr' in the list, with `pred' pointing to `curr'?
    @Override
    protected boolean valid(Node<T> pred, Node<T> curr) {
        return pred.valid() && curr.valid() && pred.next() == curr;
    }

    // `add' is as in OptimisticSet, but using LazySet's `valid'

    // first invalidate, then physically remove
    @Override
    protected boolean rawRemove(Node<T> pred, Node<T> node, int key) {
        // first, invalidate node if found
        if (node.key() == key)
            node.setInvalid();
        // the rest is as before
        return super.rawRemove(pred, node, key);
    }

    // `remove' is as in OptimisticSet, but using LazySet's `rawRemove'

    // find and validate without locking; result also depends on validity check
    // (the method's structure is identical to SequentialSearch's plus validation)
    @Override
    public boolean has(T item) {
        int key = item.hashCode();
        Node<T> pred = null, curr = null; // dummy initialization
        // first node from head whose key >= `key'
        Position<T> where = find(head, key);  // find without locking
        pred = where.pred;
        curr = where.curr;
        return curr.valid() && rawHas(curr, key);  // node not being removed and key found
    }

    // node with a `valid' mark
    protected Node<T> newNode(T item) {
        return new ValidatedNode<>(item);
    }

    // node with a `valid' mark
    protected Node<T> newNode(int key) {
        return new ValidatedNode<>(key);
    }
}
